@@ -1,7 +1,12 @@
 /*--------------------------------*-C-*---------------------------------*
  * File:	xdefaults.c
  *----------------------------------------------------------------------*
- * Copyright 1997 1998 mj olesen <olesen@me.queensu.ca>
+ * $Id: xdefaults.c,v 1.34.2.5 2000/07/05 06:51:45 gcw Exp $
+ *
+ * All portions of code are copyright by their respective author/s.
+ * Copyright (C) 1994      Robert Nation <nation@rocket.sanders.lockheed.com>
+ *				- original version
+ * Copyright (C) 1997,1998 mj olesen <olesen@me.queensu.ca>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,227 +21,248 @@
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
- *
- *----------------------------------------------------------------------*
+ *----------------------------------------------------------------------*/
+/*----------------------------------------------------------------------*
  * get resources from ~/.Xdefaults or ~/.Xresources with the memory-saving
  * default or with XGetDefault() (#define USE_XGETDEFAULT)
  *
  * Coding style:
- *	resource strings are indicated by an `rs_' prefix followed by
+ *	resource strings are indicated by an `Rs_' prefix followed by
  *	the resource name.
- *	eg, `rs_saveLines' is the resource string corresponding to
+ *	eg, `Rs_saveLines' is the resource string corresponding to
  *	    the `saveLines' resource
  *----------------------------------------------------------------------*/
 
 #include "rxvt.h"		/* NECESSARY */
+#include "version.h"
+#include "xdefaults.intpro"	/* PROTOS for internal routines */
 
 /* #define DEBUG_RESOURCES */
 
-/* local functions referenced */
-/*{{{ local variables */
-static const char *rs_loginShell = NULL;
-static const char *rs_utmpInhibit = NULL;
-static const char *rs_scrollBar = NULL;
-static const char *rs_scrollBar_right = NULL;
-static const char *rs_scrollBar_floating = NULL;
-static const char *rs_scrollTtyOutputInh = NULL;
-static const char *rs_scrollKeypress = NULL;
-
-#if defined (HOTKEY_CTRL) || defined (HOTKEY_META)
-static const char *rs_bigfont_key = NULL;
-static const char *rs_smallfont_key = NULL;
-#endif
-
-#ifndef NO_MAPALERT
-#ifdef MAPALERT_OPTION
-static const char *rs_mapAlert = NULL;
-#endif
-#endif
-static const char *rs_visualBell = NULL;
-static const char *rs_reverseVideo = NULL;
-
-#ifdef META8_OPTION
-static const char *rs_meta8 = NULL;
-#endif
-#ifdef KANJI
-static const char *rs_kanji_encoding = NULL;
-#endif
-#ifdef GREEK_SUPPORT
-static const char *rs_greek_keyboard = NULL;
-#endif
-/*}}} */
-
 /*{{{ monolithic option/resource structure: */
 /*
  * `string' options MUST have a usage argument
  * `switch' and `boolean' options have no argument
- *
  * if there's no desc(ription), it won't appear in usage()
  */
-static const struct {
-    unsigned long   flag;
-    const char    **dp;		/* data pointer */
-    const char     *const kw;	/* keyword */
-    const char     *const opt;	/* option */
-    const char     *const arg;	/* argument */
-    const char     *const desc;	/* description */
-} optList[] = {
 
 /* INFO() - descriptive information only */
-#define INFO(opt,arg,desc)       {0, NULL, NULL, opt, arg, desc}
+#define INFO(opt, arg, desc)				\
+    {0, NULL, NULL, opt, arg, desc}
+
 /* STRG() - command-line option, with/without resource */
-#define STRG(p,kw,opt,arg,desc)  {0, &p, kw, opt, arg, desc}
+#define STRG(p, kw, opt, arg, desc)			\
+    {0, &(rs[p]), kw, opt, arg, desc}
+
 /* RSTRG() - resource/long-option */
-#define RSTRG(p,kw,arg)          {0, &p, kw, NULL, arg, NULL}
+#define RSTRG(p, kw, arg)				\
+    {0, &(rs[p]), kw, NULL, arg, NULL}
+
 /* BOOL() - regular boolean `-/+' flag */
-#define BOOL(p,kw,opt,flag,desc) {(Opt_Boolean|flag), &p, kw, opt, NULL, desc}
+#define BOOL(p, kw, opt, flag, desc)			\
+    {(Opt_Boolean|flag), &(rs[p]), kw, opt, NULL, desc}
+
 /* SWCH() - `-' flag */
-#define SWCH(opt,flag,desc)      {(flag), NULL, NULL, opt, NULL, desc}
+#define SWCH(opt, flag, desc)				\
+    {(flag), NULL, NULL, opt, NULL, desc}
+
 /* convenient macros */
-#define optList_strlen(i) (optList[i].flag ? 0 : (optList[i].arg ? strlen (optList[i].arg) : 1))
-#define optList_isBool(i) (optList[i].flag & Opt_Boolean)
-#define optList_size()    (sizeof(optList)/sizeof(optList[0]))
+#define optList_strlen(i)						\
+    (optList[i].flag ? 0 : (optList[i].arg ? strlen (optList[i].arg) : 1))
+#define optList_isBool(i)						\
+    (optList[i].flag & Opt_Boolean)
+#define optList_isReverse(i)						\
+    (optList[i].flag & Opt_Reverse)
+#define optList_size()							\
+    (sizeof(optList) / sizeof(optList[0]))
 
-	STRG(display_name, NULL, "display", "displayname",
-	 "X server to contact"),
-	STRG(rs_term_name, "termname", "tn", "termName",
-	     "value of the TERM environment variable"),
-	STRG(rs_geometry, "geometry", "geometry", "geom",
-	     "size (in characters) and position"),
-	STRG(display_name, NULL, "d", NULL, NULL),	/* short form */
-	STRG(rs_geometry, NULL, "g", NULL, NULL),	/* short form */
-	BOOL(rs_reverseVideo, "reverseVideo", "rv", Opt_reverseVideo,
-	     "reverse video"),
-	STRG(rs_color[bgColor], "background", "bg", "color",
-	     "background color"),
-	STRG(rs_color[fgColor], "foreground", "fg", "color",
-	     "foreground color"),
-/* colors: command-line long-option = resource name */
-	RSTRG(rs_color[minColor + 0], "color0", "color"),
-	RSTRG(rs_color[minColor + 1], "color1", "color"),
-	RSTRG(rs_color[minColor + 2], "color2", "color"),
-	RSTRG(rs_color[minColor + 3], "color3", "color"),
-	RSTRG(rs_color[minColor + 4], "color4", "color"),
-	RSTRG(rs_color[minColor + 5], "color5", "color"),
-	RSTRG(rs_color[minColor + 6], "color6", "color"),
-	RSTRG(rs_color[minColor + 7], "color7", "color"),
-#ifndef NO_BRIGHTCOLOR
-	RSTRG(rs_color[minBright + 0], "color8", "color"),
-	RSTRG(rs_color[minBright + 1], "color9", "color"),
-	RSTRG(rs_color[minBright + 2], "color10", "color"),
-	RSTRG(rs_color[minBright + 3], "color11", "color"),
-	RSTRG(rs_color[minBright + 4], "color12", "color"),
-	RSTRG(rs_color[minBright + 5], "color13", "color"),
-	RSTRG(rs_color[minBright + 6], "color14", "color"),
-	RSTRG(rs_color[minBright + 7], "color15", "color"),
-#endif				/* NO_BRIGHTCOLOR */
-#ifndef NO_BOLDUNDERLINE
-	RSTRG(rs_color[colorBD], "colorBD", "color"),
-	RSTRG(rs_color[colorUL], "colorUL", "color"),
-#endif				/* NO_BOLDUNDERLINE */
-#ifdef KEEP_SCROLLCOLOR
-	RSTRG(rs_color[scrollColor], "scrollColor", "color"),
-#endif				/* KEEP_SCROLLCOLOR */
-#if defined (XPM_BACKGROUND) || (MENUBAR_MAX)
-	RSTRG(rs_path, "path", "search path"),
-#ifdef XPM_BACKGROUND
-	STRG(rs_backgroundPixmap,
-	     "backgroundPixmap",
-	     "pixmap", "file[;geom]", "background pixmap"),
-#endif				/* defined (XPM_BACKGROUND) || (MENUBAR_MAX) */
-#endif				/* XPM_BACKGROUND */
-#if (MENUBAR_MAX)
-	RSTRG(rs_menu, "menu", "name[;tag]"),
-#endif
-#ifndef NO_BOLDFONT
-	STRG(rs_boldFont, "boldFont", "fb", "fontname", "bold text font"),
-#endif
-	STRG(rs_font[0], "font", "fn", "fontname", "normal text font"),
-/* fonts: command-line option = resource name */
-	RSTRG(rs_font[1], "font1", "fontname"),
-	RSTRG(rs_font[2], "font2", "fontname"),
-	RSTRG(rs_font[3], "font3", "fontname"),
-	RSTRG(rs_font[4], "font4", "fontname"),
-#ifdef KANJI
-	STRG(rs_kfont[0], "kfont", "fk", "fontname", "kanji font"),
-
-/* fonts: command-line option = resource name */
-	RSTRG(rs_kfont[1], "kfont1", "fontname"),
-	RSTRG(rs_kfont[2], "kfont2", "fontname"),
-	RSTRG(rs_kfont[3], "kfont3", "fontname"),
-	RSTRG(rs_kfont[4], "kfont4", "fontname"),
-
-	STRG(rs_kanji_encoding, "kanji_encoding", "km", "mode",
-	     "kanji encoding; mode = eucj | sjis"),
-#endif				/* KANJI */
-#ifdef BIG5
-	STRG(rs_cfont[0], "cfont", "fc", "fontname", "chinese font"),
-
-/* fonts: command-line option = resource name */
-	RSTRG(rs_cfont[1], "cfont1", "fontname"),
-	RSTRG(rs_cfont[2], "cfont2", "fontname"),
-	RSTRG(rs_cfont[3], "cfont3", "fontname"),
-	RSTRG(rs_cfont[4], "cfont4", "fontname"),
-#endif				/* BIG5 */
-#ifdef GREEK_SUPPORT
-	STRG(rs_greek_keyboard, "greek_keyboard", "grk", "mode",
-	     "greek keyboard mapping; mode = iso | ibm"),
-#endif
-	SWCH("iconic", Opt_iconic, "start iconic"),
-	SWCH("ic", Opt_iconic, NULL),	/* short form */
-	STRG(rs_name, NULL, "name", "string",
-	     "client instance, icon, and title strings"),
-	STRG(rs_title, "title", "title", "string", "title name for window"),
-	STRG(rs_title, NULL, "T", NULL, NULL),	/* short form */
-	STRG(rs_iconName, "iconName", "n", "string", "icon name for window"),
-#ifndef NO_CURSORCOLOR
-	STRG(rs_color[cursorColor], "cursorColor", "cr", "color",
-	     "cursor color"),
-/* command-line option = resource name */
-	RSTRG(rs_color[cursorColor2], "cursorColor2", "color"),
-#endif				/* NO_CURSORCOLOR */
-	STRG(rs_color[pointerColor], "pointerColor", "pr", "color",
-	     "pointer color"),
-	STRG(rs_color[borderColor], "borderColor", "bd", "color",
-	     "border color"),
-	BOOL(rs_loginShell, "loginShell", "ls", Opt_loginShell, "login shell"),
-	BOOL(rs_scrollBar, "scrollBar", "sb", Opt_scrollBar, "scrollbar"),
-	BOOL(rs_scrollBar_right, "scrollBar_right", "sr", Opt_scrollBar_right,
-	     "scrollbar right"),
-	BOOL(rs_scrollBar_floating, "scrollBar_floating", "st",
-	     Opt_scrollBar_floating, "scrollbar without a trough"),
-	BOOL(rs_scrollTtyOutputInh, "scrollTtyOutput", "si",
-	     Opt_scrollTtyOutputInh, "scroll-on-tty-output inhibit"),
-	BOOL(rs_scrollKeypress, "scrollTtyKeypress", "sk",
-	     Opt_scrollKeypress, "scroll-on-keypress"),
-	STRG(rs_saveLines, "saveLines", "sl", "number",
-	     "number of scrolled lines to save"),
-	BOOL(rs_utmpInhibit, "utmpInhibit", "ut", Opt_utmpInhibit,
-	     "utmp inhibit"),
-	BOOL(rs_visualBell, "visualBell", "vb", Opt_visualBell, "visual bell"),
-#ifndef NO_MAPALERT
-#ifdef MAPALERT_OPTION
-	BOOL(rs_mapAlert, "mapAlert", NULL, Opt_mapAlert, NULL),
+static const struct {
+    const unsigned long flag;	/* Option flag */
+    const char    **dp;		/* data pointer */
+    const char     *kw;		/* keyword */
+    const char     *opt;	/* option */
+    const char     *arg;	/* argument */
+    const char     *desc;	/* description */
+} optList[] = {
+    STRG(Rs_display_name, NULL, "d", NULL, NULL),	/* short form */
+    STRG(Rs_display_name, NULL, "display", "string", "X server to contact"),
+    STRG(Rs_term_name, "termName", "tn", "string",
+         "value of the TERM environment variable"),
+    STRG(Rs_geometry, NULL, "g", NULL, NULL),	/* short form */
+    STRG(Rs_geometry, "geometry", "geometry", "geometry",
+         "size (in characters) and position"),
+    SWCH("C", Opt_console, "intercept console messages"),
+    SWCH("iconic", Opt_iconic, "start iconic"),
+    SWCH("ic", Opt_iconic, NULL),	/* short form */
+    BOOL(Rs_reverseVideo, "reverseVideo", "rv", Opt_reverseVideo,
+         "reverse video"),
+    BOOL(Rs_loginShell, "loginShell", "ls", Opt_loginShell, "login shell"),
+    BOOL(Rs_scrollBar, "scrollBar", "sb", Opt_scrollBar, "scrollbar"),
+    BOOL(Rs_scrollBar_right, "scrollBar_right", "sr", Opt_scrollBar_right,
+         "scrollbar right"),
+    BOOL(Rs_scrollBar_floating, "scrollBar_floating", "st",
+         Opt_scrollBar_floating, "scrollbar without a trough"),
+    BOOL(Rs_scrollTtyOutput, "scrollTtyOutput", NULL,
+         Opt_scrollTtyOutput, NULL),
+    BOOL(Rs_scrollTtyOutput, NULL, "si", Opt_Reverse | Opt_scrollTtyOutput,
+	 "scroll-on-tty-output inhibit"),
+    BOOL(Rs_scrollKeypress, "scrollTtyKeypress", "sk",
+         Opt_scrollKeypress, "scroll-on-keypress"),
+#ifdef TRANSPARENT
+    BOOL(Rs_transparent, "inheritPixmap", "ip",
+	 Opt_transparent, "inherit parent pixmap"),
+    SWCH("tr", Opt_transparent, NULL),
 #endif
+    BOOL(Rs_utmpInhibit, "utmpInhibit", "ut", Opt_utmpInhibit,
+         "utmp inhibit"),
+    BOOL(Rs_visualBell, "visualBell", "vb", Opt_visualBell, "visual bell"),
+#if ! defined(NO_MAPALERT) && defined(MAPALERT_OPTION)
+    BOOL(Rs_mapAlert, "mapAlert", NULL, Opt_mapAlert, NULL),
 #endif
 #ifdef META8_OPTION
-	BOOL(rs_meta8, "meta8", NULL, Opt_meta8, NULL),
+    BOOL(Rs_meta8, "meta8", NULL, Opt_meta8, NULL),
 #endif
+    STRG(Rs_color + Color_bg, "background", "bg", "color", "background color"),
+    STRG(Rs_color + Color_fg, "foreground", "fg", "color", "foreground color"),
+    RSTRG(Rs_color + minCOLOR + 0, "color0", "color"),
+    RSTRG(Rs_color + minCOLOR + 1, "color1", "color"),
+    RSTRG(Rs_color + minCOLOR + 2, "color2", "color"),
+    RSTRG(Rs_color + minCOLOR + 3, "color3", "color"),
+    RSTRG(Rs_color + minCOLOR + 4, "color4", "color"),
+    RSTRG(Rs_color + minCOLOR + 5, "color5", "color"),
+    RSTRG(Rs_color + minCOLOR + 6, "color6", "color"),
+    RSTRG(Rs_color + minCOLOR + 7, "color7", "color"),
+#ifndef NO_BRIGHTCOLOR
+    RSTRG(Rs_color + minBrightCOLOR + 0, "color8", "color"),
+    RSTRG(Rs_color + minBrightCOLOR + 1, "color9", "color"),
+    RSTRG(Rs_color + minBrightCOLOR + 2, "color10", "color"),
+    RSTRG(Rs_color + minBrightCOLOR + 3, "color11", "color"),
+    RSTRG(Rs_color + minBrightCOLOR + 4, "color12", "color"),
+    RSTRG(Rs_color + minBrightCOLOR + 5, "color13", "color"),
+    RSTRG(Rs_color + minBrightCOLOR + 6, "color14", "color"),
+    RSTRG(Rs_color + minBrightCOLOR + 7, "color15", "color"),
+#endif				/* NO_BRIGHTCOLOR */
+#ifndef NO_BOLDUNDERLINE
+    RSTRG(Rs_color + Color_BD, "colorBD", "color"),
+    RSTRG(Rs_color + Color_UL, "colorUL", "color"),
+#endif				/* NO_BOLDUNDERLINE */
+#ifdef KEEP_SCROLLCOLOR
+    RSTRG(Rs_color + Color_scroll, "scrollColor", "color"),
+    RSTRG(Rs_color + Color_trough, "troughColor", "color"),
+#endif				/* KEEP_SCROLLCOLOR */
+#if defined (XPM_BACKGROUND) || (MENUBAR_MAX)
+    RSTRG(Rs_path, "path", "search path"),
+#endif				/* defined (XPM_BACKGROUND) || (MENUBAR_MAX) */
+#ifdef XPM_BACKGROUND
+    STRG(Rs_backgroundPixmap, "backgroundPixmap",
+         "pixmap", "file[;geom]", "background pixmap"),
+#endif				/* XPM_BACKGROUND */
+#if (MENUBAR_MAX)
+    RSTRG(Rs_menu, "menu", "name[;tag]"),
+#endif
+#ifndef NO_BOLDFONT
+    STRG(Rs_boldFont, "boldFont", "fb", "fontname", "bold text font"),
+#endif
+    STRG(Rs_font, "font", "fn", "fontname", "normal text font"),
+/* fonts: command-line option = resource name */
+#if NFONTS > 1
+    RSTRG(Rs_font + 1, "font1", "fontname"),
+#endif
+#if NFONTS > 2
+    RSTRG(Rs_font + 2, "font2", "fontname"),
+#endif
+#if NFONTS > 3
+    RSTRG(Rs_font + 3, "font3", "fontname"),
+#endif
+#if NFONTS > 4
+    RSTRG(Rs_font + 4, "font4", "fontname"),
+#endif
+#if NFONTS > 5
+    RSTRG(Rs_font + 5, "font5", "fontname"),
+#endif
+#if NFONTS > 6
+    RSTRG(Rs_font + 6, "font6", "fontname"),
+#endif
+#if NFONTS > 7
+    RSTRG(Rs_font + 7, "font7", "fontname"),
+#endif
+#ifdef MULTICHAR_SET
+    STRG(Rs_mfont, "mfont", "fm", "fontname", "multichar font"),
 
+/* fonts: command-line option = resource name */
+# if NFONTS > 1
+    RSTRG(Rs_mfont + 1, "mfont1", "fontname"),
+# endif
+# if NFONTS > 2
+    RSTRG(Rs_mfont + 2, "mfont2", "fontname"),
+# endif
+# if NFONTS > 3
+    RSTRG(Rs_mfont + 3, "mfont3", "fontname"),
+# endif
+# if NFONTS > 4
+    RSTRG(Rs_mfont + 4, "mfont4", "fontname"),
+# endif
+# if NFONTS > 5
+    RSTRG(Rs_mfont + 5, "mfont5", "fontname"),
+# endif
+# if NFONTS > 6
+    RSTRG(Rs_mfont + 6, "mfont6", "fontname"),
+# endif
+# if NFONTS > 7
+    RSTRG(Rs_mfont + 7, "mfont7", "fontname"),
+# endif
+#endif				/* MULTICHAR_SET */
+#ifdef MULTICHAR_SET
+    STRG(Rs_multichar_encoding, "multichar_encoding", "km", "mode",
+         "multiple-character font encoding; mode = eucj | sjis | big5 | gb"),
+#endif				/* MULTICHAR_SET */
+#ifdef USE_XIM
+    STRG(Rs_preeditType, "preeditType", "pt", "style",
+         "input style of input method; style = OverTheSpot | OffTheSpot | Root"),
+    STRG(Rs_inputMethod, "inputMethod", "im", "name",
+         "name of input method"),
+#endif                         /* USE_XIM */
+#ifdef GREEK_SUPPORT
+    STRG(Rs_greek_keyboard, "greek_keyboard", "grk", "mode",
+         "greek keyboard mapping; mode = iso | ibm"),
+#endif
+    STRG(Rs_name, NULL, "name", "string",
+         "client instance, icon, and title strings"),
+    STRG(Rs_title, "title", "title", "string", "title name for window"),
+    STRG(Rs_title, NULL, "T", NULL, NULL),	/* short form */
+    STRG(Rs_iconName, "iconName", "n", "string", "icon name for window"),
+#ifndef NO_CURSORCOLOR
+    STRG(Rs_color + Color_cursor, "cursorColor", "cr", "color",
+         "cursor color"),
+/* command-line option = resource name */
+    RSTRG(Rs_color + Color_cursor2, "cursorColor2", "color"),
+#endif				/* NO_CURSORCOLOR */
+    STRG(Rs_color + Color_pointer, "pointerColor", "pr", "color",
+         "pointer color"),
+    STRG(Rs_color + Color_border, "borderColor", "bd", "color",
+         "border color"),
+    STRG(Rs_saveLines, "saveLines", "sl", "number",
+         "number of scrolled lines to save"),
+#ifndef NO_BACKSPACE_KEY
+    RSTRG(Rs_backspace_key, "backspacekey", "string"),
+#endif
+#ifndef NO_DELETE_KEY
+    RSTRG(Rs_delete_key, "deletekey", "string"),
+#endif
+    RSTRG(Rs_selectstyle, "selectstyle", "string"),
 #ifdef PRINTPIPE
-	RSTRG(rs_print_pipe, "print-pipe", "string"),
+    RSTRG(Rs_print_pipe, "print-pipe", "string"),
 #endif
-
 #if defined (HOTKEY_CTRL) || defined (HOTKEY_META)
-	RSTRG(rs_bigfont_key, "bigfont_key", "keysym"),
-	RSTRG(rs_smallfont_key, "smallfont_key", "keysym"),
+    RSTRG(Rs_bigfont_key, "bigfont_key", "keysym"),
+    RSTRG(Rs_smallfont_key, "smallfont_key", "keysym"),
 #endif
+    STRG(Rs_modifier, "modifier", "mod", "modifier", "alt | meta | hyper | super | mod1 | ... | mod5"),
+    INFO("xrm", "string", "X resource"),
 #ifdef CUTCHAR_RESOURCE
-	RSTRG(rs_cutchars, "cutchars", "string"),
+    RSTRG(Rs_cutchars, "cutchars", "string"),
 #endif				/* CUTCHAR_RESOURCE */
-
-	SWCH("C", Opt_console, "intercept console messages"),
-	INFO("e", "command arg ...", "command to execute")
+    INFO("e", "command arg ...", "command to execute")
 };
 
 #undef INFO
@@ -248,7 +274,7 @@ static const struct {
 
 /*{{{ usage: */
 /*----------------------------------------------------------------------*/
-/* PROTO */
+/* EXTPROTO */
 void
 usage(int type)
 {
@@ -257,10 +283,11 @@ usage(int type)
 #define INDENT 30
     fprintf(stderr, "\nUsage v%s : (", VERSION);
 #ifdef XPM_BACKGROUND
-    fprintf(stderr, "XPM,");
-#ifdef XPM_BUFFERING
+# ifdef XPM_BUFFERING
     fprintf(stderr, "XPM-buffer,");
-#endif
+# else
+    fprintf(stderr, "XPM,");
+# endif
 #endif
 #ifdef UTMP_SUPPORT
     fprintf(stderr, "utmp,");
@@ -271,23 +298,35 @@ usage(int type)
 #ifdef KANJI
     fprintf(stderr, "Kanji,");
 #endif
-#ifdef BIG5
+#ifdef ZH
     fprintf(stderr, "Chinese,");
 #endif
+#ifdef ZHCN
+    fprintf(stderr, "Chinese(GB),");
+#endif
 #ifdef XTERM_SCROLLBAR
     fprintf(stderr, "XTerm-scrollbar,");
 #endif
 #ifdef GREEK_SUPPORT
     fprintf(stderr, "Greek,");
 #endif
+#ifdef RXVT_GRAPHICS
+    fprintf(stderr, "graphics,");
+#endif
+#ifdef NO_BACKSPACE_KEY
+    fprintf(stderr, "no backspace,");
+#endif
+#ifdef NO_DELETE_KEY
+    fprintf(stderr, "no delete,");
+#endif
 #ifdef NO_RESOURCES
     fprintf(stderr, "NoResources");
 #else
-#ifdef USE_XGETDEFAULT
+# ifdef USE_XGETDEFAULT
     fprintf(stderr, "XGetDefaults");
-#else
+# else
     fprintf(stderr, ".Xdefaults");
-#endif
+# endif
 #endif
 
     fprintf(stderr, ")\n%s", APL_NAME);
@@ -356,36 +395,31 @@ usage(int type)
 
 #ifdef KEYSYM_RESOURCE
 	fprintf(stderr, "    " "keysym.sym" ": %*s\n",
-		(INDENT - strlen("keysym.sym")),
-		"keysym"
-#ifdef USE_XGETDEFAULT
-		" # (command-line only)"
-#endif
-	    );
+		(INDENT - strlen("keysym.sym")), "keysym");
 #endif
 	fprintf(stderr, "\n    -help to list options\n\n");
 	break;
     }
     exit(EXIT_FAILURE);
+    /* NOTREACHED */
 }
 /*}}} */
 
 /*{{{ get command-line options before getting resources */
-/* PROTO */
+/* EXTPROTO */
 void
-get_options(int argc, char *argv[])
+get_options(int argc, const char * const *argv)
 {
     int             i, bad_option = 0;
-    static const char *const On = "ON";
-    static const char *const Off = "OFF";
+    static const char On[3] = "ON", Off[4] = "OFF";
 
     for (i = 1; i < argc; i++) {
 	int             entry, longopt = 0;
-	const char     *flag;
-	char           *opt = argv[i];
-
+	const char     *flag, *opt;
+	
+	opt = argv[i];
 #ifdef DEBUG_RESOURCES
-	fprintf(stderr, "argv[%d] = %s: ", i, argv[i]);
+	fprintf(stderr, "argv[%d] = %s: ", i, opt);
 #endif
 	if (*opt == '-') {
 	    flag = On;
@@ -403,17 +437,21 @@ get_options(int argc, char *argv[])
 
 	if (!strcmp(opt, "help"))
 	    usage(longopt ? 2 : 1);
+	if (!strcmp(opt, "h"))
+	    usage(0);
 
     /* feature: always try to match long-options */
 	for (entry = 0; entry < optList_size(); entry++)
-	    if ((optList[entry].kw && !strcmp(opt, optList[entry].kw)) ||
-		(!longopt &&
-		 optList[entry].opt && !strcmp(opt, optList[entry].opt)))
+	    if ((optList[entry].kw && !strcmp(opt, optList[entry].kw))
+		|| (!longopt
+		    && optList[entry].opt && !strcmp(opt, optList[entry].opt)))
 		break;
 
 	if (entry < optList_size()) {
+	    if (optList_isReverse(entry))
+		flag = flag == On ? Off : On;
 	    if (optList_strlen(entry)) {	/* string value */
-		char           *str = argv[++i];
+		const char     *str = argv[++i];
 
 #ifdef DEBUG_RESOURCES
 		fprintf(stderr, "string (%s,%s) = ",
@@ -426,10 +464,10 @@ get_options(int argc, char *argv[])
 #endif
 		    *(optList[entry].dp) = str;
 
-		    /* special cases are handled in main.c:main() to allow
-                     * X resources to set these values before we settle for
-                     * default values
-                     */
+		/* special cases are handled in main.c:main() to allow
+		 * X resources to set these values before we settle for
+		 * default values
+		 */
 		}
 #ifdef DEBUG_RESOURCES
 		else
@@ -452,7 +490,7 @@ get_options(int argc, char *argv[])
 #ifdef KEYSYM_RESOURCE
 	/* if (!strncmp (opt, "keysym.", strlen ("keysym."))) */
 	if (Str_match(opt, "keysym.")) {
-	    char           *str = argv[++i];
+	    const char     *str = argv[++i];
 
 	/*
 	 * '7' is strlen("keysym.")
@@ -496,7 +534,7 @@ get_options(int argc, char *argv[])
 /*
  * a replacement for strcasecmp() to avoid linking an entire library
  */
-/* PROTO */
+/* INTPROTO */
 int
 my_strcasecmp(const char *s1, const char *s2)
 {
@@ -512,24 +550,24 @@ my_strcasecmp(const char *s1, const char *s2)
 
 /*}}} */
 
-#ifdef KEYSYM_RESOURCE
+# ifdef KEYSYM_RESOURCE
 /*
  * Define key from XrmEnumerateDatabase.
  *   quarks will be something like
  *      "rxvt" "keysym" "0xFF01"
  *   value will be a string
  */
-/* PROTO */
+/* ARGSUSED */
+/* INTPROTO */
 Bool
-define_key(XrmDatabase * database, XrmBindingList bindings, XrmQuarkList quarks, XrmRepresentation * type, XrmValue * value, XPointer closure)
+define_key(XrmDatabase *database, XrmBindingList bindings, XrmQuarkList quarks, XrmRepresentation *type, XrmValue *value, XPointer closure)
 {
     int             last;
 
     for (last = 0; quarks[last] != NULLQUARK; last++)	/* look for last quark in list */
 	;
     last--;
-    if (parse_keysym(XrmQuarkToString(quarks[last]), (char *)value->addr) == 1)
-	return True;
+    parse_keysym(XrmQuarkToString(quarks[last]), (char *)value->addr);
     return False;
 }
 
@@ -541,17 +579,18 @@ define_key(XrmDatabase * database, XrmBindingList bindings, XrmQuarkList quarks,
  *      NULL for ~/.Xdefaults and
  *      non-NULL for command-line options (need to allocate)
  */
-/* PROTO */
+#define NEWARGLIM	500	/* `reasonable' size */
+/* INTPROTO */
 int
-parse_keysym(char *str, char *arg)
+parse_keysym(const char *str, const char *arg)
 {
-    char           *key_string;
     int             n, sym;
+    char           *key_string, *newarg = NULL;
+    char            newargstr[NEWARGLIM];
 
     if (arg == NULL) {
 	if ((n = Str_match(str, "keysym.")) == 0)
 	    return 0;
-
 	str += n;		/* skip `keysym.' */
     }
 /* some scanf() have trouble with a 0x prefix */
@@ -566,72 +605,58 @@ parse_keysym(char *str, char *arg)
 		return -1;
 
 	/* cue to ':', it's there since sscanf() worked */
-	    str = strchr(str, ':');
-	    str++;
-	    arg = Str_trim(str);
-	    if (arg == NULL)
-		return -1;
+	    STRNCPY(newargstr, strchr(str, ':') + 1, NEWARGLIM - 1);
+	    newargstr[NEWARGLIM - 1] = '\0';
+	    newarg = Str_trim(newargstr);
 	}
     } else {
     /*
      * convert keysym name to keysym number
      */
+	STRNCPY(newargstr, str, NEWARGLIM - 1);
+	newargstr[NEWARGLIM - 1] = '\0';
 	if (arg == NULL) {
-	    arg = str;
-
-	    arg = strchr(str, ':');
-	    if (arg == NULL)
-		return -1;
-
-	    *arg++ = '\0';
-	    arg = Str_trim(arg);
-	    if (arg == NULL)
+	    if ((newarg = strchr(newargstr, ':')) == NULL)
 		return -1;
+	    *newarg++ = '\0';	/* terminate keysym name */
+	    (void) Str_trim(newarg);
 	}
-	sym = XStringToKeysym(str);
-
-	if (sym == None)
+	if ((sym = XStringToKeysym(newargstr)) == None)
 	    return -1;
     }
 
-/* only do extended keys */
-    if (sym >= 0xFF00)
-	sym -= 0xFF00;
-
-/* bad number or already set */
-    if (sym < 0 || sym > 0xFF || KeySym_map[sym] != NULL)
+    if (sym < 0xFF00 || sym > 0xFFFF)	/* we only do extended keys */
+	return -1;
+    sym &= 0xFF;
+    if (KeySym_map[sym] != NULL)	/* already set ? */
 	return -1;
 
-    if ((n = strlen(arg)) == 0)
-	return -1;
-
-    key_string = MALLOC((n + 2) * sizeof(char));
-
-    strcpy(key_string + 1, arg);
-
-    n = Str_escaped(key_string + 1);
-    if (n) {
-	if (n > 255)
-	    n = 255;
-	key_string[0] = n;
-	KeySym_map[sym] = key_string;
-    } else {
-	FREE(key_string);
-	return -1;
+    if (newarg == NULL) {
+	STRNCPY(newargstr, arg, NEWARGLIM - 1);
+	newargstr[NEWARGLIM - 1] = '\0';
+	newarg = newargstr;
     }
+    if (strlen(newargstr) == 0
+	|| (n = Str_escaped(newargstr)) == 0)
+	return -1;
+    MIN_IT(n, 255);
+    key_string = MALLOC((n + 1) * sizeof(char));
+    key_string[0] = n;
+    STRNCPY(key_string + 1, newarg, n);
+    KeySym_map[sym] = (unsigned char *) key_string;
 
     return 1;
 }
-#endif				/* KEYSYM_RESOURCE */
+# endif				/* KEYSYM_RESOURCE */
 
-#ifndef USE_XGETDEFAULT
+# ifndef USE_XGETDEFAULT
 /*{{{ get_xdefaults() */
 /*
  * the matching algorithm used for memory-save fake resources
  */
-/* PROTO */
+/* INTPROTO */
 void
-get_xdefaults(FILE * stream, const char *name)
+get_xdefaults(FILE *stream, const char *name)
 {
     unsigned int    len;
     char           *str, buffer[256];
@@ -645,16 +670,16 @@ get_xdefaults(FILE * stream, const char *name)
 	while (*str && isspace(*str))
 	    str++;		/* leading whitespace */
 
-	if ((str[len] != '*' && str[len] != '.') ||
-	    (len && strncmp(str, name, len)))
+	if ((str[len] != '*' && str[len] != '.')
+	    || (len && strncmp(str, name, len)))
 	    continue;
 	str += (len + 1);	/* skip `name*' or `name.' */
 
-#ifdef KEYSYM_RESOURCE
+# ifdef KEYSYM_RESOURCE
 	if (!parse_keysym(str, NULL))
-#endif				/* KEYSYM_RESOURCE */
+# endif				/* KEYSYM_RESOURCE */
 	    for (entry = 0; entry < optList_size(); entry++) {
-		const char     *const kw = optList[entry].kw;
+		const char     *kw = optList[entry].kw;
 
 		if (kw == NULL)
 		    continue;
@@ -662,17 +687,23 @@ get_xdefaults(FILE * stream, const char *name)
 		if (str[n] == ':' && Str_match(str, kw)) {
 		/* skip `keyword:' */
 		    str += (n + 1);
-		    str = Str_skip_space(str);
-		    str = Str_trim(str);
-		    n = (str ? strlen(str) : 0);
+		    (void) Str_trim(str);
+		    n = strlen(str);
 		    if (n && *(optList[entry].dp) == NULL) {
 		    /* not already set */
+			int             s;
 			char           *p = MALLOC((n + 1) * sizeof(char));
 
-			strcpy(p, str);
+			STRCPY(p, str);
 			*(optList[entry].dp) = p;
 			if (optList_isBool(entry)) {
-			    if (!my_strcasecmp(str, "TRUE"))
+			    s = my_strcasecmp(p, "TRUE") == 0
+				|| my_strcasecmp(p, "YES") == 0
+				|| my_strcasecmp(p, "ON") == 0
+				|| my_strcasecmp(p, "1") == 0;
+			    if (optList_isReverse(entry))
+				s = !s;
+			    if (s)
 				Options |= (optList[entry].flag);
 			    else
 				Options &= ~(optList[entry].flag);
@@ -685,25 +716,26 @@ get_xdefaults(FILE * stream, const char *name)
     rewind(stream);
 }
 /*}}} */
-#endif				/* ! USE_XGETDEFAULT */
+# endif				/* ! USE_XGETDEFAULT */
 #endif				/* NO_RESOURCES */
 
 /*{{{ read the resources files */
 /*
  * using XGetDefault() or the hand-rolled replacement
  */
-/* PROTO */
+/* ARGSUSED */
+/* EXTPROTO */
 void
-extract_resources(Display * display, const char *name)
+extract_resources(Display *display, const char *name)
 {
 #ifndef NO_RESOURCES
-#ifdef USE_XGETDEFAULT
+# ifdef USE_XGETDEFAULT
 /*
  * get resources using the X library function
  */
     int             entry;
 
-#ifdef XrmEnumOneLevel
+#  ifdef XrmEnumOneLevel
     XrmName         name_prefix[3];
     XrmClass        class_prefix[3];
     char           *displayResource;
@@ -724,21 +756,28 @@ extract_resources(Display * display, const char *name)
 	screenDatabase = XrmGetStringDatabase(screenResource);
     XrmMergeDatabases(screenDatabase, &database);
     XrmSetDatabase(display, database);
-#endif
+#  endif
 
     for (entry = 0; entry < optList_size(); entry++) {
+	int             s;
 	char           *p;
 	const char     *kw = optList[entry].kw;
 
 	if (kw == NULL || *(optList[entry].dp) != NULL)
 	    continue;		/* previously set */
-	if ((p = XGetDefault(display, name, kw)) != NULL ||
-	    (p = XGetDefault(display, APL_SUBCLASS, kw)) != NULL ||
-	    (p = XGetDefault(display, APL_CLASS, kw)) != NULL) {
+	if ((p = XGetDefault(display, name, kw)) != NULL
+	    || (p = XGetDefault(display, APL_SUBCLASS, kw)) != NULL
+	    || (p = XGetDefault(display, APL_CLASS, kw)) != NULL) {
 	    *optList[entry].dp = p;
 
 	    if (optList_isBool(entry)) {
-		if (!my_strcasecmp(p, "TRUE"))
+		s = my_strcasecmp(p, "TRUE") == 0
+		    || my_strcasecmp(p, "YES") == 0
+		    || my_strcasecmp(p, "ON") == 0
+		    || my_strcasecmp(p, "1") == 0;
+		if (optList_isReverse(entry))
+		    s = !s;
+		if (s)
 		    Options |= (optList[entry].flag);
 		else
 		    Options &= ~(optList[entry].flag);
@@ -749,8 +788,8 @@ extract_resources(Display * display, const char *name)
 /*
  * [R5 or later]: enumerate the resource database
  */
-#ifdef XrmEnumOneLevel
-#ifdef KEYSYM_RESOURCE
+#  ifdef XrmEnumOneLevel
+#   ifdef KEYSYM_RESOURCE
     name_prefix[0] = XrmStringToName(name);
     name_prefix[1] = XrmStringToName("keysym");
     name_prefix[2] = NULLQUARK;
@@ -773,13 +812,12 @@ extract_resources(Display * display, const char *name)
 			 XrmEnumOneLevel,
 			 define_key,
 			 NULL);
-#endif
-#endif
+#   endif
+#  endif
 
-#else				/* USE_XGETDEFAULT */
+# else				/* USE_XGETDEFAULT */
 /* get resources the hard way, but save lots of memory */
-    const char     *fname[] =
-    {".Xdefaults", ".Xresources"};
+    const char     *const fname[] = { ".Xdefaults", ".Xresources" };
     FILE           *fd = NULL;
     char           *home;
 
@@ -788,7 +826,7 @@ extract_resources(Display * display, const char *name)
 	char           *f = NULL;
 
 	for (i = 0; i < (sizeof(fname) / sizeof(fname[0])); i++) {
-	    f = REMALLOC(f, (len + strlen(fname[i])) * sizeof(char));
+	    f = REALLOC(f, (len + strlen(fname[i])) * sizeof(char));
 
 	    sprintf(f, "%s/%s", home, fname[i]);
 
@@ -822,7 +860,7 @@ extract_resources(Display * display, const char *name)
     get_xdefaults(fd, name);
     get_xdefaults(fd, APL_SUBCLASS);
 
-#ifdef XAPPLOADDIR
+#  ifdef XAPPLOADDIR
     {
 	FILE           *ad = fopen(XAPPLOADDIR "/" APL_SUBCLASS, "r");
 
@@ -831,41 +869,45 @@ extract_resources(Display * display, const char *name)
 	    fclose(ad);
 	}
     }
-#endif				/* XAPPLOADDIR */
+#  endif				/* XAPPLOADDIR */
 
     get_xdefaults(fd, APL_CLASS);
     get_xdefaults(fd, "");	/* partial match */
     if (fd != NULL)
 	fclose(fd);
-#endif				/* USE_XGETDEFAULT */
+# endif				/* USE_XGETDEFAULT */
 #endif				/* NO_RESOURCES */
 
 /*
  * even without resources, at least do this setup for command-line
  * options and command-line long options
  */
-#ifdef KANJI
-    set_kanji_encoding(rs_kanji_encoding);
+#ifdef MULTICHAR_SET
+    set_multichar_encoding(rs[Rs_multichar_encoding]);
 #endif
 #ifdef GREEK_SUPPORT
 /* this could be a function in grkelot.c */
 /* void set_greek_keyboard (const char * str); */
-    if (rs_greek_keyboard) {
-	if (!strcmp(rs_greek_keyboard, "iso"))
+    if (rs[Rs_greek_keyboard]) {
+	if (!strcmp(rs[Rs_greek_keyboard], "iso"))
 	    greek_setmode(GREEK_ELOT928);	/* former -grk9 */
-	else if (!strcmp(rs_greek_keyboard, "ibm"))
+	else if (!strcmp(rs[Rs_greek_keyboard], "ibm"))
 	    greek_setmode(GREEK_IBM437);	/* former -grk4 */
     }
 #endif				/* GREEK_SUPPORT */
 
-#define to_keysym(pks,str) do { KeySym sym;\
-if (str && ((sym = XStringToKeysym(str)) != 0)) *pks = sym; } while (0)
-
 #if defined (HOTKEY_CTRL) || defined (HOTKEY_META)
-    to_keysym(&ks_bigfont, rs_bigfont_key);
-    to_keysym(&ks_smallfont, rs_smallfont_key);
+    {
+	KeySym          sym;
+
+	if (rs[Rs_bigfont_key]
+	    && ((sym = XStringToKeysym(rs[Rs_bigfont_key])) != 0))
+	    ks_bigfont = sym;
+	if (rs[Rs_smallfont_key]
+	    && ((sym = XStringToKeysym(rs[Rs_smallfont_key])) != 0))
+	    ks_smallfont = sym;
+    }
 #endif
-#undef to_keysym
 }
 /*}}} */
 /*----------------------- end-of-file (C source) -----------------------*/
